package big.epfl.ch.doppler.imageprocessing;

import static big.epfl.ch.doppler.objects.Color.BLACK;
import static big.epfl.ch.doppler.objects.Color.WHITE;
import static big.epfl.ch.doppler.objects.Color.blue;
import static big.epfl.ch.doppler.objects.Color.color;
import static big.epfl.ch.doppler.objects.Color.green;
import static big.epfl.ch.doppler.objects.Color.hue;
import static big.epfl.ch.doppler.objects.Color.red;
import static java.lang.Math.abs;
import ij.process.EllipseFitter;
import ij.process.FloatProcessor;
import ij.process.ImageProcessor;

import java.awt.Polygon;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import big.epfl.ch.doppler.objects.Image;
import big.epfl.ch.doppler.objects.Point;
import big.epfl.ch.doppler.preprocessing.Window;
import big.epfl.ch.doppler.snake.Esnake_ImageJ_;

/**
 * @author Axel Vandebrouck & Brune Bastide
 *
 */
public class BloodFlow implements Runnable {

	private Map<Integer, Float> colorMap;
	private Image img;

	private Esnake_ImageJ_ snake;
	private boolean usingSnake;

	private float flow = 0;
	private float r = 0;
	private float angle = 0;
	private int vesselarea = 0;
	

	/**
	 * @param img : The image from which we want to compute the blood flow 
	 * @param colorMap : Mapping the color of a pixel with 
	 * 					 the corresponding punctual speed of the blood
	 * @param usingSnake : A boolean which indicates if we use snakes to compute the blood flow or not
	 */
	public BloodFlow(Image img, Map<Integer, Float> colorMap, boolean usingSnake){
		this.img = img;
		this.colorMap = new HashMap<>(colorMap);
		this.snake = new Esnake_ImageJ_();
		this.usingSnake = usingSnake;
		this.r = ellipticCorrection(img);
	}

	public BloodFlow(){
	}

	/**
	 * @param img : The original image
	 * @return a new image containing only the pixels representing the blood in the original image 
	 * 		in gray scale values by applying the following formula for each pixel :
	 * 		newImage(x, y) = |red(img(x, y)-green(img(x, y))| + |blue(img(x, y)-green(img(x, y))| 
	 */
	public Image extractBloodChannels(Image img){
		int[] blood = new int[img.size()];
		for(int i = 0; i < img.size(); i++) {
			blood[i] = color((int)(abs(red(img.pixels[i])-green(img.pixels[i]))+abs(blue(img.pixels[i])-green(img.pixels[i]))));
		}
		return new Image(img.width, img.height, blood);
	}
	
	/**
	 * @param img : The original image
	 * @return a new image containing only the pixels representing the blood in the original image
	 */
	public Image getBlood(Image img){
		Image binaryImg = getBinaryBlood(img);
		int[] blood = new int[img.size()];
		for(int i=0; i<blood.length; i++){
			blood[i] = (binaryImg.pixels[i] == WHITE) ? img.pixels[i] : BLACK;
		}
		return new Image(img.width, img.height, blood);
	}
	
	/**
	 * @param img : The original image
	 * @return a new image obtained by thresholding the original image : 
	 * 		the pixels representing the blood will be white and the others black 
	 */
	public Image getBinaryBlood(Image img){
		return extractBloodChannels(img).threshold(0.2f);
	}

	/**
	 * @return the blood flow
	 */
	public float getFlow(){
		return flow;
	}

	/**
	 * @return the points of the snakes
	 */
	public Point[] getSnakePoints(){
		return snake.getSnakePoints();
	}

	/**
	 * @param img : The original image
	 * @return the computation of the blood flow depending if we are using snakes or not
	 */
	private float computeFlow(Image img){

		Image blood = getBinaryBlood(getLCC(img, false));
		if(usingSnake){
			snake.run(new FloatProcessor(blood.width, blood.height, blood.pixels));
			List<Point> snakePoints = Arrays.asList(snake.getSnakePoints());
			int area = 0, red=0, green=0, blue=0;

			for(int i = 0; i<img.size(); i++){
				if(new Point(i%img.width, i/img.width).isInside(snakePoints)){
					if(blood.pixels[i] != BLACK){
						area++;
						red += red(img.pixels[i]);
						green += green(img.pixels[i]);
						blue += blue(img.pixels[i]);
					}
				}
			}
			return (area != 0) ? area*getSpeed(color(red/area, green/area, blue/area))*r : 0;
		}
		else {
			int totalFlow = 0;
			for(int i = 0; i<blood.size(); i++){
				if(blood.pixels[i] == WHITE){
					vesselarea++;
					totalFlow += getSpeed(img.pixels[i]);
				}
			}
			return totalFlow*r;		
		}
	}

	/**
	 * @param color : the color to be mapped in order to get the corresponding speed
	 * @return the speed associated with the color. <b>PI or PI/2 ??</br>
	 * 
	 */
	private float getSpeed(int color){

		if(color == BLACK){
			return 0;
		}
		else {
			float speed = 0;
			if(colorMap.containsKey(color)) {
				speed = colorMap.get(color);
			}
			else {
				float minDelta = Integer.MAX_VALUE;
				float h = hue(color);
				for(int c : colorMap.keySet()){
					float delta = Math.abs(h-hue(c));
					if(delta < minDelta){
						minDelta = delta;
						speed = colorMap.get(c);
					}
				}
			}
			return speed/(float)(Math.sin(Math.PI/2-Math.acos(r))); //corrected for PI/2, correct??
		}
	}
	
	/**@author raoul
	 * @return incidence: between US ray and vessel.
	 * Caution! They calculated 90-incidence instead!
	 */
	private float computeAngle(){
		double incidence = (Math.PI/2-Math.acos(r));
		return (float) Math.toDegrees(incidence);
	}
	
	public float getAngle(){
		return angle;
	}

	/**
	 * @param img : The original image
	 * @return a new image containing only the largest connected component of the image, 
	 * 		i.e the largest spot representing the blood
	 */
	public Image getLCC(Image img, boolean isBinary){
		Image binaryImg = isBinary ? img : getBinaryBlood(img);
		int label = 1;
		Map<Integer, Integer> eqIDs = new HashMap<Integer, Integer>(); 
		int[] cc = new int[img.size()];
		for(int i = 0; i < cc.length; i++){
			if(binaryImg.pixels[i] == WHITE){
				int[] neighboors = {cc[i-img.width-1], cc[i-img.width], cc[i-img.width+1], cc[i-1]};
				if(cc[i-1] > 0 || cc[i-img.width-1] > 0 || cc[i-img.width] > 0 || cc[i-img.width+1] > 0){
					int min = label;
					for(int n = 0; n<neighboors.length; n++){
						if(neighboors[n] > 0 && neighboors[n] < min){
							min = neighboors[n];
						}
					}
					cc[i] = min;
					for(int n = 0; n<neighboors.length; n++){
						if(neighboors[n] > cc[i] && eqIDs.get(neighboors[n]) > cc[i]){
							eqIDs.put(neighboors[n], cc[i]);
						}
					}
				}
				else cc[i] = label++;
				
				if(!eqIDs.containsKey(cc[i])){
					eqIDs.put(cc[i], cc[i]);
				}
			}
			else cc[i] = 0;
		}
		Map<Integer, Integer> componentsSize = new HashMap<Integer, Integer>(); 
		for(int i = 0; i < cc.length; i++){
			if(cc[i] > 0){
				while(eqIDs.get(cc[i]) < cc[i]){
					cc[i] = eqIDs.get(cc[i]);
				}
				if(componentsSize.containsKey(cc[i])){
					componentsSize.put(cc[i], componentsSize.get(cc[i])+1);
				}
				else componentsSize.put(cc[i], 1);
			}
		}
		int maxSize=0, maxComponentID=-1;
		for(int id : componentsSize.keySet()){
			if(componentsSize.get(id) > maxSize){
				maxSize = componentsSize.get(id);
				maxComponentID = id;
			}
		}
		int[] lcc = new int[img.size()];
		for(int i = 0; i<lcc.length; i++){
			lcc[i] = (cc[i] == maxComponentID ? img.pixels[i] : BLACK);
		}
		return new Image(img.width, img.height, lcc);
	}

	/**
	 * @param img : The original image in black and white
	 * @return the quotient of the small axis length by the big axis length of the best ellipse fitting the blood
	 */
	public float ellipticCorrection(Image img){	
		Image blood = getBinaryBlood(getLCC(img, false));
		EllipseFitter e = new EllipseFitter();
		ImageProcessor ip = new FloatProcessor(blood.width, blood.height, blood.pixels);
		ip.setRoi(usingSnake ? snake.getPolygon() : bloodPolygon(blood));
		e.fit(ip, null);
		double r = e.minor/e.major;
		return (float)(Double.isNaN(r) ? 0 : r);
	}

	/**
	 * @param img : The original image in black and white
	 * @return the polygon constructed with the border of the blood
	 */
	public Polygon bloodPolygon(Image img){
		List<Point> gradient = new ArrayList<Point>();
		for(int i=0; i<img.size(); i++){
			if(isBorderPixel(img, i)){
				gradient.add(new Point(i%img.width, i/img.width));
			}
		}
		if(gradient.isEmpty()){
			return null;
		}
		else {
			Point origin = gradient.get(0);
			float distMax = 0;
			int centerX=0, centerY=0;
			for(int i=1; i<gradient.size(); i++){
				float dist = origin.dist(gradient.get(i));
				if(dist > distMax){
					distMax = dist;
					centerX = (origin.getX()+gradient.get(i).getX())/2;
					centerY = (origin.getY()+gradient.get(i).getY())/2;
				}
			}
			List<Point> sortedCorners = new Window(gradient).sortCorners(new Point(centerX, centerY));
			int nPoints = sortedCorners.size();
			int[] xPoints = new int[nPoints], yPoints = new int[nPoints];
			for(int i = 0; i<nPoints; i++){
				xPoints[i] = sortedCorners.get(i).getX();
				yPoints[i] = sortedCorners.get(i).getY();
			}
			return new Polygon(xPoints, yPoints, nPoints);
		}
	}

	/**
	 * @param img : The original image in black and white
	 * @param index : the index of the pixel we want to check 
	 * @return true if the pixel indicated by index is a border pixel of the blood
	 */
	private boolean isBorderPixel(Image img, int index){
		return img.pixels[index] == WHITE 
				&& !(img.pixels[index-1]==WHITE && img.pixels[index+1]==WHITE 
				&& img.pixels[index-img.width]==WHITE && img.pixels[index+img.width]==WHITE);
	}

	public int getVesselarea() {
		//float area = (float) (Math.pow(r*Math.sqrt(vesselarea/(Math.PI*r))/2, 2)*Math.PI);
		return vesselarea;
	}
	
	public float getSpeed() {
		return (float) ((flow*Math.sin(Math.PI/2-Math.acos(r)))/(r*vesselarea));
	}

	@Override
	public void run() {
		flow = computeFlow(img);
		angle = computeAngle();
		
//		float flow_epf = (Math.round(100*flow/6400))/100f;
//		float n = (float) Math.sqrt(4*flow/flow_epf);
//		System.out.println(angle);
//		System.out.println("recalculated flow: "+(4*flow)/(160*160));
	}
}
