package big.epfl.ch.doppler.imageprocessing;

import static big.epfl.ch.doppler.objects.Color.mean;

import java.util.List;

import big.epfl.ch.doppler.objects.Image;
import big.epfl.ch.doppler.objects.Point;
import big.epfl.ch.doppler.preprocessing.EdgeDetection;
import big.epfl.ch.doppler.preprocessing.Window;

/**
 * @author Axel Vandebrouck & Brune Bastide 
 *
 */
public class Deinterlacing {

	private EdgeDetection edgeDetection;
	private Window window;
	private List<Point> corners;

	/**
	 * @param img : the original image from which we extract the corners of the blue parallelogram
	 * @throws IllegalArgumentException if the blue parallelogram is not detected
	 */
	public Deinterlacing(Image img) throws IllegalArgumentException {
		edgeDetection = new EdgeDetection();
		window = new Window(edgeDetection.getIntersections(edgeDetection.pipeline(img)));
		corners = window.sortCorners(new Point(img.width/2, img.height/2));
		boolean isOut = false;
		for(Point p : corners){
			if(p.getX()<0 || p.getX()>img.width || p.getY()<0 || p.getY()>img.height){
				isOut = true;
			}
		}
		if(isOut) throw new IllegalArgumentException("The detected corners are out of the image");
	}
	
	/**
	 * @param notSortedCorners : the corners of the blue parallelogram selected by hand
	 * @param center : the center of the image
	 * @throws IllegalArgumentException if the numbers of corners is not equal to 4
	 */
	public Deinterlacing(List<Point> notSortedCorners, Point center) throws IllegalArgumentException {
		if(notSortedCorners.size() != 4){
			throw new IllegalArgumentException("The number of corners is not good : " + notSortedCorners.size() + " instead of 4");
		}
		window = new Window(notSortedCorners);
		corners = window.sortCorners(center);
	}
	
	/**
	 * @return the corners of the blue parallelogram
	 */
	public List<Point> getCorners(){
		return corners;
	}
	
	/**
	 * @param img : the original image from the movie to be deinterlaced 
	 * @return four new images of size w/4 x h/4 (with w and h are the dimensions of the input image)
	 * 		which are deinterlaced 
	 */
	public Image[] deinterlace(Image img){
				
		Image[] newImgs = new Image[4];
		for(int k = 0; k<newImgs.length; k++){
			newImgs[k] = new Image(img.width/4, img.height/4);
		}
		int i, j;

		for(int y = 0; y<newImgs[0].height; y++){
			for(int x = 0; x<newImgs[0].width; x++){
				Point p = new Point(4*x, 4*y);
				if(p.isInside(corners)){
					i = y*newImgs[0].width+x;
					j = 16*y*newImgs[0].width+4*x;
					for(int k=0; k<newImgs.length; k++){
						newImgs[k].pixels[i] = mean(mean(img.pixels[j+k*img.width], img.pixels[j+k*img.width+1]), 
								mean(img.pixels[j+k*img.width+2], img.pixels[j+k*img.width+3]));
					}
				}
			}
		}		
		return newImgs;
	}
}