package big.epfl.ch.doppler.objects;

import java.io.IOException;
import java.util.Arrays;
import java.util.List;
import java.util.Map;

import static big.epfl.ch.doppler.gui.Test.MAX_FLOW;
import big.epfl.ch.doppler.imageprocessing.BloodFlow;
import big.epfl.ch.doppler.imageprocessing.ConfigurationFileReader;

/**
 * @author Axel Vandebrouck & Brune Bastide
 *
 */
public class Color {
	
	/**
	 * The white color
	 */
	public static final int WHITE = color(255);
	/**
	 * The black color
	 */
	public static final int BLACK = color(0);
		
	/**
	 * @param argb : the value of a pixel
	 * @return the red component of the pixel
	 */
	public static float red(int argb){
		return (argb >> 16) & 0xFF;
	}
	
	/**
	 * @param argb : the value of a pixel
	 * @return the green component of the pixel
	 */
	public static float green(int argb){
		return (argb >> 8) & 0xFF;
	}
	
	/**
	 * @param argb : the value of a pixel
	 * @return the blue component of the pixel
	 */
	public static float blue(int argb){
		return argb & 0xFF;
	}
	
	/**
	 * @param argb : the value of a pixel
	 * @return the gray scale value of the pixel
	 */
	public static float gray(int argb){
		return (int)((red(argb)+green(argb)+blue(argb))/3);
	}
	
	/**
	 * @param gray : the gray component of the color between 0 and 255
	 * @return an integer representing the gray scale color from the given value in parameter
	 */
	public static int color(int gray){
		return color(gray, gray, gray);
	}
	
	/**
	 * @param r : the red component of the color between 0 and 255
	 * @param g : the green component of the color between 0 and 255
	 * @param b : the blue component of the color between 0 and 255
	 * @return an integer representing the RGB color from the given values in parameter
	 * 		constructed by shifting the different parameters, each one on 8 bits
	 */
	public static int color(int r, int g, int b){
		return (r << 16) | (g << 8) | b;
	}
	
	/**
	 * @param rgb : an integer representing a RGB color
	 * @return the hue parameter of the HSB conversion from the RGB color (between 0 and 360)
	 */
	public static float hue(int rgb){
		float r = red(rgb)/255, g = green(rgb)/255, b = blue(rgb)/255;
		float Cmax = max(r, g, b);
		float Cmin = min(r, g, b);
		float delta = Cmax-Cmin;
		int h = delta != 0  ? (int)(60*(Cmax == r ? (g-b)/delta : Cmax == g ? 2+(b-r)/delta : 4+(r-g)/delta)) : 0;
		return h<0 ? h+360 : h;
	}
	
	/**
	 * @param rgb : an integer representing a RGB color
	 * @return the saturation parameter of the HSB conversion from the RGB color (between 0 and 100)
	 */
	public static float saturation(int rgb){
		float r = red(rgb)/255, g = green(rgb)/255, b = blue(rgb)/255;
		float max = max(r, g, b);
		float min = min(r, g, b);
		float delta = max-min;
		return max != 0 ? (int)(100*delta/max) : 0;
	}
	
	/**
	 * @param rgb : an integer representing a RGB color
	 * @return the brightness parameter of the HSB conversion from the RGB color (between 0 and 100)
	 */
	public static float brightness(int rgb){
		return max(red(rgb), green(rgb), blue(rgb))*100/255;
	}
	
	/**
	 * @param img : An image
	 * @return the saturation image of the given image in parameter 
	 */
	public static Image getSaturationImage(Image img){
		Image newImg = new Image(img.width, img.height);
		for(int i = 0; i<img.width*img.height; i++){
			newImg.pixels[i] = color(Math.round(saturation(img.pixels[i]))*255/100);
		}
		return newImg;
	}
	
	/**
	 * @param rgb1 : an integer representing a RGB color
	 * @param rgb2 : an integer representing a RGB color
	 * @return the mean of the two colors given in parameters
	 */
	public static int mean(int rgb1, int rgb2){
		int r = (int)(red(rgb1)+red(rgb2))/2;
		int g = (int)(green(rgb1)+green(rgb2))/2;
		int b = (int)(blue(rgb1)+blue(rgb2))/2;
		return color(r, g, b);
	}
	
	/**
	 * @param colorMap : Mapping a color with the corresponding speed of the blood 
	 * @param max : the max speed of the blood 
	 * @throws IOException if a mistake happens while reading the configuration file 
	 * 		containing the values of the colors of the color scale 
	 */
	public static void fillColorMap(Map<Integer, Float> colorMap, float max) throws IOException {
		List<Integer> colors = new ConfigurationFileReader("").read();
		float step = 2*max/(colors.size()-1);
		for(int i=0; i<colors.size(); i++){
			colorMap.put(colors.get(i), max-i*step);
		}
	}
	
	/**
	 * @param colorMap : the colorMap to be filled 
	 * @param img : the original image from which we extract the color scale to fill the color map 
	 * @param corners : the corners of the blue parallelogram 
	 */
	public static void fillColorMap(Map<Integer, Float> colorMap, Image img, List<Point> corners) {
		
		BloodFlow bf = new BloodFlow();
		Image scaleChannels = bf.extractBloodChannels(img);
		
		for(int i = 0; i<scaleChannels.size(); i++){
			if(new Point(i%img.width, i/img.width).isInside(corners)) 
				scaleChannels.pixels[i] = BLACK;
		}
		Image binaryScale = bf.getLCC(scaleChannels.threshold(0.23f), true);
		int Xmin=img.width, Xmax=0, Ymin=img.height, Ymax=0;
		for(int i=0; i<binaryScale.size(); i++){
			if(binaryScale.pixels[i]==WHITE){
				int x = i%img.width, y=i/img.width;
				Xmin = min(x, Xmin);
				Xmax = max(x, Xmax);
				Ymin = min(y, Ymin);
				Ymax = max(y, Ymax);
			}					
		}
		final float max = MAX_FLOW;
		float step = 2*max/(Ymax-Ymin);
		int x = (Xmin+Xmax)/2;
		for(int i=Ymin*img.width+x; i<=Ymax*img.width+x; i+=img.width){
			int n = (i-x)/img.width-Ymin;
			colorMap.put(img.pixels[i], max - n*step);
		}
	}
	
	/**
	 * @param colorMap : the colorMap to be filled 
	 * @param img : the original image from which we extract the color scale to fill the color map 
	 * @param corners : the corners of the blue parallelogram 
	 */
	public static void fillColorMap(float threshold, Map<Integer, Float> colorMap, Image img, List<Point> corners) {
		
		BloodFlow bf = new BloodFlow();
		Image scaleChannels = bf.extractBloodChannels(img);
		
		for(int i = 0; i<scaleChannels.size(); i++){
			if(new Point(i%img.width, i/img.width).isInside(corners)) 
				scaleChannels.pixels[i] = BLACK;
		}
		Image binaryScale = bf.getLCC(scaleChannels.threshold(threshold), true);
		int Xmin=img.width, Xmax=0, Ymin=img.height, Ymax=0;
		for(int i=0; i<binaryScale.size(); i++){
			if(binaryScale.pixels[i]==WHITE){
				int x = i%img.width, y=i/img.width;
				Xmin = min(x, Xmin);
				Xmax = max(x, Xmax);
				Ymin = min(y, Ymin);
				Ymax = max(y, Ymax);
			}					
		}
		final float max = MAX_FLOW;
		float step = 2*max/(Ymax-Ymin);
		int x = (Xmin+Xmax)/2;
		for(int i=Ymin*img.width+x; i<=Ymax*img.width+x; i+=img.width){
			int n = (i-x)/img.width-Ymin;
			colorMap.put(img.pixels[i], max - n*step);
		}
	}
	
	/**
	 * @author raoul
	 * @param colorMap : the colorMap to be filled 
	 * @param img : the original image from which we extract the color scale to fill the color map 
	 * @param corners : the corners of the blue parallelogram 
	 * @return the colormap with a custom max speed
	 */
	public static void fillColorMap(Map<Integer, Float> colorMap, Image img, List<Point> corners, float maxspeed) {
		
		BloodFlow bf = new BloodFlow();
		Image scaleChannels = bf.extractBloodChannels(img);
		
		for(int i = 0; i<scaleChannels.size(); i++){
			if(new Point(i%img.width, i/img.width).isInside(corners)) 
				scaleChannels.pixels[i] = BLACK;
		}
		Image binaryScale = bf.getLCC(scaleChannels.threshold(0.23f), true);
		int Xmin=img.width, Xmax=0, Ymin=img.height, Ymax=0;
		for(int i=0; i<binaryScale.size(); i++){
			if(binaryScale.pixels[i]==WHITE){
				int x = i%img.width, y=i/img.width;
				Xmin = min(x, Xmin);
				Xmax = max(x, Xmax);
				Ymin = min(y, Ymin);
				Ymax = max(y, Ymax);
			}					
		}
		final float max = maxspeed;
		float step = 2*max/(Ymax-Ymin);
		int x = (Xmin+Xmax)/2;
		for(int i=Ymin*img.width+x; i<=Ymax*img.width+x; i+=img.width){
			int n = (i-x)/img.width-Ymin;
			colorMap.put(img.pixels[i], max - n*step);
		}
	}
	
	/**
	 * @param r : the red component of the color between 0 and 255
	 * @param g : the green component of the color between 0 and 255
	 * @param b : the blue component of the color between 0 and 255
	 * @return the max between the three values
	 */
	private static float max(float r, float g, float b){
		return r > g ? (r > b ? r : b) : (g > b ? g : b);
	}
	
	/**
	 * @param r : the red component of the color between 0 and 255
	 * @param g : the green component of the color between 0 and 255
	 * @param b : the blue component of the color between 0 and 255
	 * @return the min between the three values
	 */
	private static float min(float r, float g, float b){
		return r < g ? (r < b ? r : b) : (g < b ? g : b);
	}
	
	/**
	 * @param a : a float number
	 * @param b : a float number
	 * @return the maximum between 'a' and 'b'
	 */
	private static int max(int a, int b){
		return a > b ? a : b;
	}
	
	/**
	 * @param a : a float number
	 * @param b : a float number
	 * @return the minimum between 'a' and 'b'
	 */
	private static int min(int a, int b){
		return a < b ? a : b;
	}
}
