package big.epfl.ch.doppler.objects;

import static big.epfl.ch.doppler.objects.Color.WHITE;
import static big.epfl.ch.doppler.objects.Color.BLACK;

/**
 * @author Axel Vandebrouck & Brune Bastide
 *
 */
public class Image {

	public int width, height;
	public int[] pixels;
	
	/**
	 * @param w : the width of the image
	 * @param h : the height of the image
	 * @throws IllegalArgumentException if at least one of the two dimensions is negative
	 */
	public Image(int w, int h) throws IllegalArgumentException {
		if(w <= 0 || h <= 0){
			throw new IllegalArgumentException("width and height must be positive");
		}
		this.width = w;
		this.height = h;
		this.pixels = new int[w*h];
	}
	
	/**
	 * @param w : the width of the image
	 * @param h : the height of the image
	 * @param pixels : the pixels of the image 
	 * @throws IllegalArgumentException if at least one of the two dimensions is negative 
	 * 		or if the size of the pixels array does not match with the given dimensions
	 */
	public Image(int w, int h, int[] pixels) throws IllegalArgumentException {
		this(w, h);
		if(pixels.length != w*h){
			throw new IllegalArgumentException("the size of the array pixels does not match with the size of the image");
		}
		setPixels(pixels);
	}

	/**
	 * @param newPixels : the new array of pixels for the image
	 * @throws IllegalArgumentException if the array of new pixels does not have the same size than the actual one 
	 */
	public void setPixels(int[] newPixels) throws IllegalArgumentException {
		if(newPixels.length != pixels.length){
			throw new IllegalArgumentException("the size of the new array of pixels does not match with the actual one");
		}
		for(int i = 0; i<pixels.length; i++){
			this.pixels[i] = newPixels[i];
		}
	}
	
	/**
	 * @param threshold : parameter to threshold the image between 0 and 1
	 * @return a new binary image thresholded with the given parameter
	 */
	public Image threshold(float threshold) throws IllegalArgumentException {
		if(threshold < 0 || threshold > 1){
			throw new IllegalArgumentException("The threshold must be between 0 and 1");
		}
		int[] newPixels = new int[size()];
		for(int i = 0; i<newPixels.length; i++){
			newPixels[i] = (Color.gray(this.pixels[i]) > 255*threshold) ? WHITE : BLACK;
		}
		return new Image(this.width, this.height, newPixels);
	}
	
	/**
	 * @param x : the x location of the upper left corner of the part of the image we want to select
	 * @param y : the y location of the upper left corner of the part of the image we want to select
	 * @param w : the width of the part of the image we want to select
	 * @param h : the height of the part of the image we want to select
	 * @return a new Image constructed from a given part of an other image 
	 * @throws IllegalArgumentException
	 */
	public Image select(int x, int y, int w, int h) throws IllegalArgumentException {
		if(x<0 || y<0 || w<0 || h<0 || x+w >this.width || y+h >this.height){
			throw new IllegalArgumentException("Selection impossible : bad parameters");
		}
		Image newImg = new Image(w, h);
		for(int i=0; i<newImg.size(); i++){
			newImg.pixels[i] = this.pixels[(y+i/newImg.width)*this.width + x+i%newImg.width];
		}
		return newImg;
	}
	
	/**
	 * @return the size of the image i.e the number of pixels it contains
	 */
	public int size(){
		return width*height;
	}
}