package big.epfl.ch.doppler.preprocessing;

import static big.epfl.ch.doppler.objects.Color.brightness;
import static big.epfl.ch.doppler.objects.Color.color;
import static big.epfl.ch.doppler.objects.Color.hue;
import static big.epfl.ch.doppler.objects.Color.saturation;
import static java.lang.Math.PI;
import static java.lang.Math.cos;
import static java.lang.Math.round;
import static java.lang.Math.sin;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import big.epfl.ch.doppler.objects.Image;
import big.epfl.ch.doppler.objects.Point;
import big.epfl.ch.doppler.objects.Vector;

/**
 * @author Axel Vandebrouck & Brune Bastide
 *
 */
public class EdgeDetection {

	// hue is between O and 360
	private final double minHue = 140; 
	private final double maxHue = 170; 

	// saturation is between O and 100
	private final double minSaturation = 0;
	private final double maxSaturation = 70;
	
	// brightness is between O and 100
	private final double minBrightness = 55; 
	private final double maxBrightness = 80;

	private final float discretizationStepsPhi = 0.06f; 
	private final float discretizationStepsR = 2.5f;
	private final int phiDim = (int)round(PI / discretizationStepsPhi);

	/**
	 * @param img : the original image 
	 * @return a list containing the polar coordinates of the detected lines in the image
	 */
	public ArrayList<Vector> pipeline(Image img){
		return hough(color2binary(img), 4);
	}

	/**
	 * @param img : the original image
	 * @return a new image in binary to detect the green parallelogram 
	 */
	public Image color2binary(Image img){

		Image newImg = new Image(img.width, img.height);
		for(int i = 0; i < img.size(); i++) {
			if((hue(img.pixels[i]) < maxHue && hue(img.pixels[i]) > minHue)
					&& saturation(img.pixels[i]) < maxSaturation && saturation(img.pixels[i]) > minSaturation
					&& (brightness(img.pixels[i]) < maxBrightness && brightness(img.pixels[i]) > minBrightness)){
				newImg.pixels[i] = color(255);
			}
			else newImg.pixels[i] = color(0);
		}
		return newImg;
	}

	/**
	 * @param edgeImg : the image in binary with the edges in white
	 * @param nLines : the number of lines we want to detect
	 * @return a list containing the polar coordinates of the detected lines in the image
	 */
	public ArrayList<Vector> hough(Image edgeImg, int nLines) {

		int rDim = round(((edgeImg.width + edgeImg.height) * 2 + 1) / discretizationStepsR);
		int[] accumulator = new int[(phiDim + 2) * (rDim + 2)];

		float r = 0;
		int accPhi = 0;
		int accR = 0;
		int idx = 0;

		// our accumulator (with a 1 pix margin around)
		// Fill the accumulator: on edge points (ie, white pixels of the edge image), 
		//store all possible (r, phi) pairs describing lines going through the point.
		for (int y = 0; y < edgeImg.height; y++) {
			for (int x = 0; x < edgeImg.width; x++) {
				// Are we on an edge?
				if (brightness(edgeImg.pixels[y * edgeImg.width + x]) != 0) {
					// ...determine here all the lines (r, phi) passing through
					// pixel (x,y), convert (r,phi) to coordinates in the
					// accumulator, and increment accordingly the accumulator.
					for (float phi = 0; phi < PI; phi += discretizationStepsPhi) {
						r = (float)(x * cos(phi) + y * sin(phi));
						accPhi = round(phi / discretizationStepsPhi);
						accR = round(r / discretizationStepsR + (rDim - 1) * 0.5f);
						idx = (accPhi + 1) * (rDim + 2) + accR + 1;
						accumulator[idx]++;
					}
				} 
			}
		}
		ArrayList<Integer> bestCandidates = new ArrayList<>();
		ArrayList<Vector> bestVectors = new ArrayList<>();
		// size of the region we search for a local maximum
		int neighbourhood = 10;
		int minVotes = 100;
		// only search around lines with more that this amount of votes // (to be adapted to your image)
		for (accR = 0; accR < rDim; accR++) {
			for (accPhi = 0; accPhi < phiDim; accPhi++) {
				// compute current index in the accumulator
				idx = (accPhi + 1) * (rDim + 2) + accR + 1; 
				if (accumulator[idx] > minVotes) {
					boolean bestCandidate = true;
					// iterate over the neighbourhood
					for(int dPhi=-neighbourhood/2; dPhi < neighbourhood/2+1; dPhi++) { 
						// check we are not outside the image
						if(accPhi+dPhi < 0 || accPhi+dPhi >= phiDim) continue; 
						for(int dR=-neighbourhood/2; dR < neighbourhood/2 +1; dR++) {
							// check we are not outside the image
							if(accR+dR < 0 || accR+dR >= rDim) continue;
							int neighbourIdx = (accPhi + dPhi + 1) * (rDim + 2) + accR + dR + 1;
							if(accumulator[idx] < accumulator[neighbourIdx]) { 
								// the current idx is not a local maximum! 
								bestCandidate=false;
								break;
							} 
						}
						if(!bestCandidate) break; 
					}
					if(bestCandidate) {
						// the current idx *is* a local maximum
						bestCandidates.add(idx);
					}
				} 
			}
		}
		Collections.sort(bestCandidates, new HoughComparator(accumulator));
		for(int i = 0; i < nLines && i < bestCandidates.size(); i++){
			idx = bestCandidates.get(i);
			// compute back the (r, phi) polar coordinates:
			accPhi = round(idx / (rDim + 2)) - 1;
			accR = idx - (accPhi + 1) * (rDim + 2) - 1;
			r = (accR - (rDim - 1) * 0.5f) * discretizationStepsR; 
			float phi = accPhi * discretizationStepsPhi;
			bestVectors.add(new Vector(r, phi));
		}
		return bestVectors;
	}

	/**
	 * @param lines : the lines obtained with the Hough algorithm
	 * @return the intersections of the lines i.e the corners of the parallelogram 
	 */
	public ArrayList<Point> getIntersections(List<Vector> lines) { 
		
		ArrayList<Point> intersections = new ArrayList<Point>(); 
		for (int i = 0; i < lines.size() - 1; i++) {
			Vector line1 = lines.get(i);
			for (int j = i + 1; j < lines.size(); j++) {
				Vector line2 = lines.get(j);
				float r1 = line1.x, r2 = line2.x;
				float phi1 = line1.y, phi2 = line2.y;
				float d = (float) (cos(phi2)*sin(phi1) - cos(phi1)*sin(phi2));
				if(d!=0){
					int x = (int) ((r2*sin(phi1) - r1*sin(phi2)) / d);
					int y = (int) ((-r2*cos(phi1) + r1*cos(phi2)) / d);
					intersections.add(new Point(x, y));
				}
			}
		}
		return intersections; 
	}

	class HoughComparator implements Comparator<Integer> { 

		int[] accumulator;
		public HoughComparator(int[] accumulator) {
			this.accumulator = accumulator; 
		}
		@Override
		public int compare(Integer l1, Integer l2) { 
			return (accumulator[l1] > accumulator[l2] 
					|| (accumulator[l1] == accumulator[l2] && l1 < l2)) ?
							-1 : 1;
		} 
	}
}