# -*- coding: utf-8 -*-
#!/usr/bin/env python

"""
This script produces test images for mvn_doppler
1. angle validation: series of ellipses of progressive axes ratio
2. color validation: series of disks covering the tested color spectrum
3. area validation: series of ellipse of progressive axes ratio
    with constant area
"""
import cv2
import math
import numpy as np
import subprocess as sp
from noise import pnoise2
from random import randint

BIGWIDTH = 720
BIGHEIGHT = 576
WIDTH = BIGWIDTH/4
HEIGHT = BIGHEIGHT/4
GREEN_COLOR = (144, 159, 119)
TEST_BIGAREA = 16000
TEST_AREA = 1000
TEX_NOISE_H = 0.85
BIGTEX_NOISE_H = TEX_NOISE_H / 4
TEX_NOISE_W = 0.6
BIGTEX_NOISE_W = TEX_NOISE_W / 4
GEOM_NOISE_H = 0.04
BIGGEOM_NOISE_H = GEOM_NOISE_H / 4
GEOM_NOISE_W = 0.01
BIGGEOM_NOISE_W = GEOM_NOISE_W / 4

# GET COLORSCALE
cap = cv2.VideoCapture('./resources/artery_src_for_dual.avi')
_, f = cap.read()

# cut and deinterlace scale region
scale = f[60:140, 120:140]
deinterlaced = scale[0::4, 0::4]
deinterlaced = deinterlaced[1:,1:]
TEST_COLOR = (int(deinterlaced[5, 2, 0]), int(deinterlaced[5, 2, 1]),
              int(deinterlaced[5, 2, 2]))
# FFMPEG WRITER (because OpenCV does not work...)
#ffmpeg -loop 1 -r 25 -i angletest_5.jpg -t 00:00:05 -vcodec mpeg4 -an angletest.mp4
def ffmpeg_write(filepath):
    command = [ 'ffmpeg',
               '-y',
               '-loop', '1',
               '-r', '25',
               '-i', filepath+'.jpg',
               '-t', '00:00:10',
               '-c:v', 'libx264',
               '-an',
               filepath+'.mp4']
    sp.call(command)

def bg_texture(height, width, wnoise, hnoise):
    img = np.zeros((height, width))  # PAL format
    for h in xrange(height):
        for w in xrange(width):
            x = pnoise2(w * wnoise, h * hnoise,
                        octaves=1,
                        persistence=0.5,
                        lacunarity=2.0,
                        base=0)
            img[h, w] = x
    return img

def bg_geom(height, width, wnoise, hnoise, base):
    img = np.zeros((height, width))  # PAL format
    for h in xrange(height):
        for w in xrange(width):
            x = pnoise2(w * wnoise, h * hnoise,
                        octaves=1,
                        persistence=0.5,
                        lacunarity=2.0,
                        base=base)
            img[h, w] = x
    return img
    
def make_bg(height, width):
    
    if height == HEIGHT and width == WIDTH:
        g = bg_geom(height, width,
                    GEOM_NOISE_W, GEOM_NOISE_H,
                    randint(0, 200))
        t = bg_texture(height, width,
                       TEX_NOISE_W, TEX_NOISE_H)

    elif height == BIGHEIGHT and width == BIGWIDTH:
        g = bg_geom(height, width,
                    BIGGEOM_NOISE_W, BIGGEOM_NOISE_H,
                    randint(0, 500))
        t = bg_texture(height, width,
                       BIGTEX_NOISE_W, BIGTEX_NOISE_H)

    else:
        raise ValueError('''incorrect parameters passed to make_bg():
                            HEIGHT = {}, WIDTH = {}'''.format(height == BIGHEIGHT,
                                                                width == BIGWIDTH))
    
    g = g + abs(g.min())
    g = g * 1/g.max()
    t = t + abs(t.min())
    t = t * 1/t.max()
    
    img = g * t
    img_equ = cv2.equalizeHist((img * 255).astype(np.uint8))
    img_blur = cv2.GaussianBlur(img_equ, (3, 3), 1, 1)
    return cv2.cvtColor((img_blur).astype(np.float32), cv2.COLOR_GRAY2RGB)
    

def interlace(img):
    int_img = np.zeros((img.shape[0] * 4, img.shape[1] * 4, 3))
    for i in xrange(0, 4):
        for x in xrange(img.shape[0]):
            for y in xrange(img.shape[1]):
                int_img[4*x, 4*y + i] = img[x, y]
    for y in xrange(0, int_img.shape[0]-4, 4):
        row_diff = (int_img[y + 4, :, :] - int_img[y, :, :])/4
        for i in xrange(1, 4):
            int_img[y + i, :, :] = int_img[y, :, :] + i * row_diff
    return int_img

#%%
# ANGLE VALIDATION
print 'Building angle images...'
d = 48
# interlaced test images             
for n in xrange(0, 90, 5):
    alpha = math.radians(90-n)
    D = d/math.cos(math.pi/2-alpha)
    img = make_bg(HEIGHT, WIDTH)
    img[5:24, 5:9] = deinterlaced  # paste the colorscale
    img = cv2.rectangle(img, (15, 5), (170, 139), GREEN_COLOR, 1)  # ROI
    img = cv2.ellipse(img, (WIDTH/2, HEIGHT/2), (d/9, int(round(D/9))),
                      0, 0, 360, TEST_COLOR, -1)
    img = interlace(img)
    filepath = './resources/angletest/interlaced/angletest_'+str(
    int(round(math.degrees(alpha))))
    print '  Building {}, with D = {}'.format(filepath+'.jpg', int(round(D/9)))
    cv2.imwrite(filepath+'.jpg', img)

# non-interlaced test images
for n in xrange(0, 90, 5):
    alpha = math.radians(90-n)
    D = d/math.cos(math.pi/2-alpha)
#    pdb.set_trace()
    img = make_bg(BIGHEIGHT, BIGWIDTH)
    img[60:140, 120:140] = scale  # paste the colorscale
    img = cv2.rectangle(img, (160, 10), (710, 566), GREEN_COLOR, 2)  # ROI
    img = cv2.ellipse(img, (BIGWIDTH/2+70, BIGHEIGHT/2), (d/2, int(round(D/2))),
                      0, 0, 360, TEST_COLOR, -1)
    filepath = './resources/angletest/noninterlaced/angletest_'+str(
    int(round(math.degrees(alpha))))
    print '  Building {}, with D = {}'.format(filepath+'.jpg', int(round(D/2)))
    cv2.imwrite(filepath+'.jpg', img)
#    #ffmpeg_write(filepath)

#%%
# COLOR VALIDATION
print 'Building color images...'
# make circles r = 250 px with the 19 color samples for the 19 height pixels

# interlaced test images
for n in xrange(0, 19):
    color = (int(deinterlaced[n, 2, 0]),
             int(deinterlaced[n, 2, 1]), int(deinterlaced[n, 2, 2]))
    img = make_bg(HEIGHT, WIDTH)
    img[5:24, 5:9] = deinterlaced  # paste the colorscale
    img = cv2.rectangle(img, (15, 5), (170, 139), GREEN_COLOR, 1)  # ROI
    #img = cv2.circle(img, (WIDTH/2+70, HEIGHT/2), 50, color, -1)
    img = cv2.ellipse(img, (WIDTH/2, HEIGHT/2), (d/2, d),
                      0, 0, 360, color, -1)
    img = interlace(img)
    filepath = './resources/colortest/interlaced/colortest_'+str(n)
    print '  Building {} with color = {}'.format(filepath+'.jpg', str(color))    
    cv2.imwrite(filepath+'.jpg', img)
#    #ffmpeg_write(filepath)

# non-interlaced test images
for n in xrange(0, 19):
    color = (int(deinterlaced[n, 2, 0]),
             int(deinterlaced[n, 2, 1]), int(deinterlaced[n, 2, 2]))
    img = make_bg(BIGHEIGHT, BIGWIDTH)
    img[60:140, 120:140] = scale  # paste the colorscale
    img = cv2.rectangle(img, (160, 10), (710, 566), GREEN_COLOR, 2)  # ROI
    #img = cv2.circle(img, (WIDTH/2+70, HEIGHT/2), 50, color, -1)
    img = cv2.ellipse(img, (BIGWIDTH/2+70, BIGHEIGHT/2), (d * 2, d * 4),
                      0, 0, 360, color, -1)
    filepath = './resources/colortest/noninterlaced/colortest_'+str(n)
    print '  Building {} with color = {}'.format(filepath+'.jpg', str(color))    
    cv2.imwrite(filepath+'.jpg', img)

#%%
# AREA VALIDATION
print 'Building area images...'

# interlaced test images
for n in xrange(0, 90, 5):
    alpha = math.radians(90-n)
    quotient = math.cos(math.pi/2-alpha)
    D = int(round(math.sqrt(TEST_AREA/(math.pi*quotient))))
    d = int(round(D*quotient))
    img = make_bg(HEIGHT, WIDTH)  # PAL format
    img[5:24, 5:9] = deinterlaced  # paste the colorscale
    img = cv2.rectangle(img, (15, 5), (170, 139), GREEN_COLOR, 1)  # ROI
    img = cv2.ellipse(img, (WIDTH/2, HEIGHT/2), (d, D),
                      0, 0, 360, TEST_COLOR, -1)
    img = interlace(img)
    filepath = './resources/areatest/interlaced/areatest_'+str(
    int(round(math.degrees(alpha))))
    print '  Building {}, with d = {} and D = {}'.format(filepath+'.jpg', d, D)
    cv2.imwrite(filepath+'.jpg', img)
#    #ffmpeg_write(filepath)

# non-interlaced test images
for n in xrange(0, 90, 5):
    alpha = math.radians(90-n)
    quotient = math.cos(math.pi/2-alpha)
    D = int(round(math.sqrt(TEST_BIGAREA/(math.pi*quotient))))
    d = int(round(D*quotient))
    img = make_bg(BIGHEIGHT, BIGWIDTH)
    img[60:140, 120:140] = scale  # paste the colorscale
    img = cv2.rectangle(img, (160, 10), (710, 566), GREEN_COLOR, 2)  # ROI
    img = cv2.ellipse(img, (BIGWIDTH/2+70, BIGHEIGHT/2), (d, D),
                      0, 0, 360, TEST_COLOR, -1)
    filepath = './resources/areatest/noninterlaced/areatest_'+str(
    int(round(math.degrees(alpha))))
    print '  Building {}, with d = {} and D = {}'.format(filepath+'.jpg', d, D)
    cv2.imwrite(filepath+'.jpg', img)
